import os

from a2ml.api.utils import fsclient
from .deploy import ModelDeploy
from ..cloud.pipeline import AugerPipelineApi
from ..cloud.endpoint import AugerEndpointApi
from ..cloud.endpoint_pipeline import AugerEndpointPipelineApi


class ModelUndeploy(object):
    """Undeploy Model locally or from Auger Cloud."""

    def __init__(self, ctx, project):
        super(ModelUndeploy, self).__init__()
        self.project = project
        self.ctx = ctx

    def execute(self, model_id, locally=False):
        if locally:
            model_path, model_zip_path = ModelDeploy(self.ctx, self.project).get_local_model_paths(model_id)
            self.ctx.log("Undeploy model. Remove local model: %s" % model_path)

            fsclient.remove_file(model_zip_path)
            fsclient.remove_folder(model_path)
        else:
            pipeline_api = AugerPipelineApi(self.ctx, None, model_id)
            if pipeline_api.has_endpoint():
                endpoint_api = AugerEndpointApi(self.ctx, None, pipeline_api.object_id)
                endpoint_props = endpoint_api.properties()
                endpoint_pipelines = sorted(endpoint_props.get('endpoint_pipelines', []), key=lambda k: k['created_at'])
                # if endpoint_pipelines and endpoint_pipelines[0]['pipeline_id'] == model_id:
                #     self.ctx.log("Undeploy Review endpoint and all models.")
                #     for pipeline in endpoint_pipelines:
                #         AugerPipelineApi(self.ctx, None, pipeline.get('pipeline_id')).remove(pipeline.get('pipeline_id'))

                #     endpoint_api.delete()

                # else:
                self.ctx.log("Undeploy model and remove from Review endpoint.")
                for pipeline in endpoint_pipelines:
                    if pipeline.get('pipeline_id') == model_id:
                        AugerPipelineApi(self.ctx, None, pipeline.get('pipeline_id')).remove(pipeline.get('pipeline_id'))

                        AugerEndpointPipelineApi(self.ctx, pipeline.get('id')).delete()
                        
                        break
            else:
                pipeline_api.remove(model_id)                    




from .base import AugerBaseApi
from ..exceptions import AugerException


class AugerClusterTaskApi(AugerBaseApi):
    """Auger Cluster Task Api."""

    def __init__(self, ctx, project_api=None,
        cluster_task_name=None, cluster_task_id=None):
        super(AugerClusterTaskApi, self).__init__(
            ctx, project_api, cluster_task_name, cluster_task_id)

    def create(self, task_args):
        return self._call_create({
            'name': self.object_name,
            'project_id': self.parent_api.object_id,
            'args_encoded': task_args},
            ['pending', 'received', 'started', 'retry']).get('result')



from .impl.cloud.rest_api import RestApi

# Patch request_list list because it returns generator, which is read-once
# so Celery tasks can't read result
old_request_list = RestApi.request_list

def new_request_list(*args):
    return list(old_request_list(*args))

RestApi.request_list = new_request_list



from .base import AugerBaseApi
from ..exceptions import AugerException


class AugerReviewAlertApi(AugerBaseApi):
    """Auger Review Alert API."""

    def __init__(self, ctx, endpoint_api, alert_id=None):
        super(AugerReviewAlertApi, self).__init__(ctx, endpoint_api, None, alert_id)

    def create(self, params):
        return self._call_create(params)

    def create_update(self, parameters=None):
        alerts = self.list()

        config = self.ctx.config
        if not parameters:
            parameters = {}

        params = {
            'endpoint_id': self.parent_api.object_id,
            'active': parameters.get('active', config.get('review/alert/active')),
            'kind': parameters.get('type', config.get('review/alert/type')),
            'threshold': float(parameters.get('threshold', config.get('review/alert/threshold'))),
            'threshold_policy': parameters.get('threshold_policy', config.get('review/alert/threshold_policy')),
            'sensitivity': int(parameters.get('sensitivity', config.get('review/alert/sensitivity'))),
            'actions': parameters.get('action', config.get('review/alert/action')),
            'notifications': parameters.get('notification', config.get('review/alert/notification'))
        }

        if params['actions'] == 'no':
            params['actions'] = 'no_actions'
        if params['notifications'] == 'no':
            params['notifications'] = 'no_notifications'

        if not alerts:
            res = self.create(params)
        else:
            params_changed = {}
            for key, value in params.items():
                alert_value = alerts[0].get(key)
                if value is not None and value != alert_value:
                    params_changed[key] = value

            if params_changed:
                params_changed['id'] = alerts[0].get('id')
                self._call_update(params_changed)



from .impl.cloud.rest_api import RestApi
from .impl.decorators import with_project
from a2ml.api.utils.decorators import error_handler, authenticated
from .impl.model import Model
from .credentials import Credentials

class AugerModel(object):

    def __init__(self, ctx):
        self.ctx = ctx
        self.credentials = Credentials(ctx).load()
        self.ctx.rest_api = RestApi(
            self.credentials.api_url, self.credentials.token)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def list(self, project, endpoints=False):
        res = []

        for model in iter(Model(self.ctx, project).list(endpoints)):
            res.append(model)
                    
        return {'models': res}

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def deploy(self, project, model_id, locally, review, name, algorithm, score, data_path, metadata=None):
        model_id = Model(self.ctx, project).deploy(model_id, locally, review, name, algorithm, score, data_path, metadata)
        return {'model_id': model_id}

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def redeploy_predict(self, project, filename, model_id, threshold, locally, data, columns, predicted_at, output, 
        no_features_in_result, score, score_true_data, predict_labels):
        from .impl.mparts.deploy import ModelDeploy
        ModelDeploy(self.ctx, project).deploy_model_locally(model_id, 
            review=False, name=None, data_path=None, locally=locally, redeploy=True)

        return self.predict(filename, model_id, threshold, locally, data, columns, predicted_at, output, 
            no_features_in_result, score, score_true_data, predict_labels, from_redeploy=True)

    @error_handler
    @authenticated
    #@with_project(autocreate=False)
    def predict(self, filename, model_id, threshold, locally, data, columns, predicted_at, output, 
        no_features_in_result, score, score_true_data, predict_labels, from_redeploy=False):
        if locally:
            self.deploy(model_id, locally, review=False, name=None, algorithm=None, score=None, data_path=None)

        try:    
            predicted = Model(self.ctx, project=None).predict(
                filename, model_id, threshold, locally, data, columns, predicted_at, output, 
                no_features_in_result, score, score_true_data, predict_labels)
        except Exception as e:
            if locally and not from_redeploy and 'Auger model error' in str(e):            
                self.redeploy_predict(filename, model_id, threshold, locally, data, columns, predicted_at, output, 
                    no_features_in_result, score, score_true_data, predict_labels)
            else:
                raise

        if output:
            self.ctx.log('Predictions stored in %s' % predicted)

        if isinstance(predicted, dict) and 'predicted' in predicted:
            return predicted

        return {'predicted': predicted}

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def actuals(self, project, model_id, filename=None, data=None, columns=None, actuals_at=None, 
        actual_date_column=None, experiment_params=None, locally=False):
        return Model(self.ctx, project).actuals(model_id, filename, data, columns, actuals_at, actual_date_column, experiment_params, locally)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def delete_actuals(self, project, model_id, with_predictions=False, begin_date=None, end_date=None, locally=False):
        return Model(self.ctx, project).delete_actuals(model_id, with_predictions, begin_date, end_date, locally)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def review_alert(self, project, model_id, parameters, name):
        return Model(self.ctx, project).review_alert(model_id, parameters, name)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def build_review_data(self, project, model_id, locally, output):
        return Model(self.ctx, project).build_review_data(model_id, locally, output)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def review(self, project, model_id):
        return Model(self.ctx, project).review(model_id)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def undeploy(self, project, model_id, locally):
        Model(self.ctx, project).undeploy(model_id, locally)
        return {'model_id': model_id}


    @error_handler
    @authenticated
    #@with_project(autocreate=False)
    def get_info(self, model_id, locally):
        return Model(self.ctx, project=None).get_info(model_id, locally)

    @error_handler
    @authenticated
    #@with_project(autocreate=False)
    def update(self, model_id, metadata, locally):
        return Model(self.ctx, project=None).update(model_id, metadata, locally)



import click
from a2ml.api.a2ml import A2ML
from a2ml.api.utils.context import pass_context


@click.command('deploy', short_help='Deploy trained model.')
@click.argument('model-id', required=False, type=click.STRING)
@click.option('--provider', '-p', type=click.Choice(['auger','azure','external']), required=False,
    help='Cloud AutoML Provider.')
@click.option('--locally', '-l', is_flag=True, default=False,
    help='Download and deploy trained model locally.')
@click.option('--review', is_flag=True, default=False,
    help='Do support model review based on actual data.')
@click.option('--name', '-n', required=False, type=click.STRING, 
    help='Model friendly name.Used as name for Review Endpoint')
@click.option('--algorithm', '-a', required=False, type=click.STRING, 
    help='Monitored model(external provider) algorithm name.')
@click.option('--score', '-s', required=False, type=float, 
    help='Monitored model(external provider) score.')
@click.option('--data-path', '-d', type=click.STRING, required=False,
    help='Data path to fit model when deploy. Return new deployed model-id')
@pass_context
def cmdl(ctx, provider, model_id, locally, review, name, algorithm, score, data_path):
    """Deploy trained model."""
    ctx.setup_logger(format='')
    A2ML(ctx, provider).deploy(model_id, locally, review, name=name, algorithm=algorithm, score=score, data_path=data_path)



import os
import codecs
import sys

from setuptools import setup
from setuptools.command.install import install

from a2ml import __version__

# Get the long description from the README file
here = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


class VerifyVersionCommand(install):
    """Verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG', '')

        if not tag.endswith(__version__, 1):
            info = "Git tag: {0} doesn't match version of a2ml: {1}".format(
                tag, __version__
            )
            sys.exit(info)


install_requires = [
    'numpy==1.23.5',
    'pandas==1.5.0',
    'joblib',
    'ruamel.yaml',
    'pyarrow',
    'scipy==1.9.1',
    'asyncio',
    'boto3',
    'auger-hub-api-client==0.7.4',
    'click', #celery 5
    'shortuuid',
    'docutils<0.16,>=0.10',
    'requests',
    'smart_open==6.2.0',
    'jsonpickle',
    'websockets',
    'liac-arff==2.4.0',
    'xlrd==1.2.0'
]

extras = {
    'testing': [
        'flake8<=3.7.9,>=3.1.0',  # version for azure
        'mock',
        'pytest',
        'pytest-cov',
        'pytest-runner',
        'pytest-xdist',
        'twine',
        'vcrpy',
        'wheel>=0.30.0,<0.31.0',
    ],
    'docs': [
        'sphinx'
    ],
    'server': [
        'aioredis',
        'celery==5.2.7',
        'fastapi==0.85',
        'gevent',
        'redis',
        's3fs>=0.4.0,<0.5.0',
        'uvicorn',
        'scikit-learn==1.2.0'
    ],
    'azure': [
        #'scikit-learn~=0.22.2',
        #'xgboost<=0.90',
        # https://github.com/Azure/azure-sdk-for-python/issues/13871
        #'azure-mgmt-resource==10.2.0',
        #this needs to move to setup.azure.py and do not include default
        'azureml-sdk[automl]==1.29.0'
    ],
    'google': [
        'google-cloud-automl'
    ],
    'predict': [
        'auger.ai.predict[all]==1.0.110'
    ],
    'predict_no_cat_lgbm': [
        'auger.ai.predict[no_cat_lgbm]==1.0.110'
    ],
    'predict_no_lgbm': [
        'auger.ai.predict[no_cat_lgbm]==1.0.110',
        'catboost'
    ]    
}

# Meta dependency groups.
all_deps = []
for group_name in extras:
    if group_name != 'predict' and group_name != 'google' and \
        group_name != 'azure' and group_name != 'predict_no_cat_lgbm' and group_name != 'predict_no_lgbm':
        all_deps += extras[group_name]

extras['all'] = all_deps


setup(
    install_requires=install_requires,
    extras_require=extras,
    tests_require=extras['testing'],
    cmdclass={
        'verify': VerifyVersionCommand
    }
)



import os
import urllib.parse
import urllib.request

from .base import AugerBaseApi
from ..exceptions import AugerException


class AugerPipelineFileApi(AugerBaseApi):
    """Auger Pipeline File API."""

    def __init__(self, ctx, experiment_api, pipeline_file_id=None):
        super(AugerPipelineFileApi, self).__init__(
            ctx, experiment_api, None, pipeline_file_id)

    def create(self, trial_id, redeploy=False):
        #TODO: suppport redeploy
        return self._call_create({'trial_id': trial_id},
            ['not_requested', 'pending'])

    def download(self, url, path_to_download, trial_id):
        if self.object_id is None:
            raise AugerException('Please provide Auger Pipeline File id')

        if not os.path.exists(path_to_download):
            os.makedirs(path_to_download)
        basename = os.path.basename(
            urllib.parse.urlparse(url).path).replace('export_','model-')
        file_name = os.path.join(path_to_download, basename)
        urllib.request.urlretrieve(url, file_name)

        return file_name

    def _get_status_name(self):
        return 'signed_s3_model_path_status'

    def _log_status(self, status):
        if status is None:
            return
        message = {
            'not_requested': 'Starting to buid model for download...',
            'pending': 'Buiding model for download...',
            'success': 'Model is ready for download...'}
        self.ctx.log(message.get(status, status))
